class ZipFileGenerator
  # Initialize with the directory to zip and the location of the output archive.
  def initialize(input_dir, output_file)
    @input_dir = input_dir
    @output_file = output_file
  end

  # Zip the input directory.
  def write
    entries = Dir.entries(@input_dir) - %w(. ..)

    ::Zip::File.open(@output_file, ::Zip::File::CREATE) do |zipfile|
      write_entries entries, '', zipfile
    end
  end

  private

  # A helper method to make the recursion work.
  def write_entries(entries, path, zipfile)
    entries.each do |e|
      zipfile_path = path == '' ? e : File.join(path, e)
      disk_file_path = File.join(@input_dir, zipfile_path)
      puts "Deflating #{disk_file_path}"

      if File.directory? disk_file_path
        recursively_deflate_directory(disk_file_path, zipfile, zipfile_path)
      else
        put_into_archive(disk_file_path, zipfile, zipfile_path)
      end
    end
  end

  def recursively_deflate_directory(disk_file_path, zipfile, zipfile_path)
    zipfile.mkdir zipfile_path
    subdir = Dir.entries(disk_file_path) - %w(. ..)
    write_entries subdir, zipfile_path, zipfile
  end

  def put_into_archive(disk_file_path, zipfile, zipfile_path)
    zipfile.get_output_stream(zipfile_path) do |f|
      file = File.open(disk_file_path, 'rb')
      f.write(file.read)
      file.close
    end
  end
end

def root_files
  %w(.env.sample .env config.ru Gemfile Gemfile.lock config.txt.sample config.txt package.json)
end

desc "Backup current code to a zip file"
task :backup do
  backup_folder = File.expand_path(File.join(__dir__, '..', 'storage', 'backups'))
  FileUtils.mkdir_p(backup_folder) unless File.exists?(backup_folder)
  zip_file = File.expand_path(backup_folder, "#{Time.now.strftime("%Y_%m_%d_%H_%M_%S")}.zip")
  root_folder = File.expand_path(File.join(__dir__, '..'))

  Dir.mktmpdir do |dir|
    FileUtils.cp_r(File.join(root_folder, 'src'), dir)
    FileUtils.cp_r(File.join(root_folder, 'vendor'), dir)
    FileUtils.cp_r(File.join(root_folder, 'rakelib'), dir)
    FileUtils.cp(root_files.map { |f| File.join(root_folder, f) }, dir)
    FileUtils.cp(File.join(root_folder, '.env'), dir)
    FileUtils.cp(File.join(root_folder, 'config.txt'), dir)

    FileUtils.rm_r(zip_file) if File.exists? zip_file
    ZipFileGenerator.new(dir, zip_file).write
  end
  zip_file
end

desc "Produce an installable package for the current version"
task :build do
  require 'tmpdir'
  require 'zip'
  require 'json'

  sh "bundle package" do |out|
    p out
  end

  version = nil
  File.open(File.join(__dir__, '../package.json')) do |f|
    version = JSON.load(f, nil, create_additions: false, symbolize_names: true)[:version]
  end

  root_folder = File.expand_path(File.join(__dir__, '..'))

  Dir.mktmpdir do |dir|
    FileUtils.cp_r(File.join(root_folder, 'src'), dir)
    FileUtils.cp_r(File.join(root_folder, 'vendor'), dir)
    FileUtils.cp_r(File.join(root_folder, 'rakelib'), dir)
    FileUtils.cp(root_files.map { |f| File.join(root_folder, f) }, dir)
    FileUtils.mkdir(File.join(dir, "log"))

    zip_file = File.join(root_folder, "rx_universe-api.#{version}.zip")
    FileUtils.rm_r(zip_file) if File.exists? zip_file
    ZipFileGenerator.new(dir, zip_file).write
  end
end


desc "Upload a previously-built release to the update server"
task :upload, [:version] do |_, args|
  require 'faraday'
  version = args[:version]
  raise "You must provide a version to release" unless version
  zip_file = "rx_universe-api.#{version}.zip"
  raise "#{zip_file} does not exist" unless File.exist?(zip_file)
  raise "No update server set" unless ENV['UPDATE_SERVER']

  conn = Faraday.new(ENV['UPDATE_SERVER']) do |f|
    f.request :multipart
  end
  p 'Uploading...'
  conn.post(
      "upload",
      file: Faraday::FilePart.new(zip_file.to_s, 'application/zip')
  )
end

desc "Upload current code to the update server"
task :publish_current => %w(build) do
  version = JSON.parse(File.read(File.join(__dir__, '..', 'package.json')))['version']
  Rake::Task["upload"].invoke(version)
end