require 'grape'
require 'grape-swagger'
require 'grape-entity'
require 'grape-swagger-entity'
require 'dotenv'
require 'grape_logging'
require 'faraday'
require_relative './cobol_runner'


def get_file(name)
  File.join(File.expand_path(__dir__), name)
end

env_file = get_file('../.env')

unless File.exists?(env_file)
  p ".env is missing. Please copy .env.sample and modify it to match your installation"
  exit 1
end

unless File.exists?(get_file("../config.txt"))
  p "config.txt is missing. Please copy config.txt.sample and modify to point to your data files"
  exit 1
end

Dotenv.load(env_file)

def resolve_type(type)
  {
      integer: Integer,
      float: Float,
      decimal: Float,
      boolean: 'boolean',
      string: String,
      date: DateTime
  }[type&.downcase&.to_sym]
end

def get_module_name(strings)
  strings = [strings] unless strings.is_a? Array
  strings.map(&:to_s).map do |str|
    str.strip.gsub(/[^a-zA-Z0-9]/, '')
  end.select do |str|
    str && str != ''
  end.map do |str|
    str.slice(0, 1).capitalize + str.slice(1..-1)
  end.join('')
end

def parse_parameter_definition(mapping, parameter_definition)
  parameter_definition.merge({
                                 type: resolve_type(parameter_definition[:type]),
                                 documentation: mapping[:method].downcase != 'get' ? {param_type: 'body'} : {}
                             }).reject { |key| [:name, :required, :file, :value].include?(key) }
end

def parse_response_entity(entity_name, entity_members)
  entity_name = get_module_name(entity_name)
  return Object.const_get(entity_name) if const_defined? entity_name
  return nil unless entity_members
  doc_class = Class.new(Grape::Entity)
  Object.const_set(entity_name, doc_class)
  doc_class.class_exec(entity_members) do
    entity_members.each do |member_def|
      if member_def[:type].downcase == 'object'
        expose member_def[:name],
               as: member_def[:name],
               using: parse_response_entity(member_def[:name], member_def[:members]),
               documentation: {is_array: !!member_def[:array]}
      else
        expose member_def[:name],
               as: member_def[:name],
               documentation: {
                   type: resolve_type(member_def[:type])&.to_s&.downcase,
                   desc: member_def[:desc],
               }
      end
    end
  end
  doc_class
end

module Rxuniverse
  class API < Grape::API
    format :json
    logger Logger.new(get_file("../log/#{ENV['RACK_ENV'] || 'production'}.log"), 'daily')
    logger.formatter = GrapeLogging::Formatters::Default.new
    use GrapeLogging::Middleware::RequestLogger, {logger: logger}

    rescue_from :all do |e|
      API.logger.error(e.message)
      error!(e, 500)
    end

    class << self
      def library_location
        ENV['OMICS6']
      end

      def mappings_location
        ENV['MAPPINGS']
      end

      def security_definitions
        needs_authentication? ? {
            "oauth2": {
                "type": "oauth2",
                "authorizationUrl": "/oauth2/authorize",
                "tokenUrl": "/oauth2/token",
                "refreshUrl": "/oauth2/refresh",
                "flow": "authorizationCode",
                "scopes": {}
            }
        } : {}
      end

      def security_object
        needs_authentication? ? [{oauth2: []}] : []
      end

      def oauth2_instructions
        needs_authentication? ? <<-SWAGGER

In addition to the listed endpoints, The Rx-Universe™ LMS API supports three additional routes for the <a href="https://tools.ietf.org/html/rfc6749#section-1.3.1">OAuth2 AuthorizationCode flow</a>: 
*  `GET` <a href="/oauth2/authorize">/oauth2/authorize?redirect_url=your_url</a> (serves the login page)
    *  On a successful login, redirect_url will be called with the parameter `?code=grant_token`
*  `POST /oauth2/token {"code": "your_grant_token"}`, which will exchange the grant token for an access token and a refresh token
   ```{
   {
      "access_token": "your_access_token", // send this in the Authorization header in future requests
      "token_type": "Bearer", // standard Oauth2 declaration; optionally prefix the Authorization header with "Bearer"
      "expires_in": 600, // time (in seconds) before the access token expires
      "refresh_token": "your_refresh_token" // provide this to /oauth2/refresh to obtain a new access token when expires_in has elapsed
   }
*  `POST /oauth2/refresh {"code": "your_refresh_token"}`, which will exchange the refresh token for a new access token and a new refresh token
   Same response structure as `POST /oauth2/token`


Use the Authorize button (right, down) to authenticate with Rx-Universe for the duration of your Swagger UI session
            SWAGGER
            : ''
      end

      def route_definitions
        @r_defs ||= JSON.parse(File.read(mappings_location), symbolize_names: true)
      end

      def cobol
        @cobol ||= CobolRunner.new("./config.txt", library_location, API.logger)
      end

      def needs_authentication?
        !!(ENV['USE_AUTH'] && !%w[false f no n].include?(ENV['USE_AUTH'].downcase))
      end

      def extract_parameter_value(definition, params)
        if definition[:type].downcase == 'literal'
          definition[:value]
        else
          params[definition[:name]]
        end
      end
    end

    before do
      header 'X-Clacks-Overhead', 'GNU Terry Pratchett'
      if Rxuniverse::API.needs_authentication? && !route.match?("/swagger") && !route.match?("/info") && !route.match?("/gui")
        token = (headers['Authorization'] || '').gsub(/bearer|token/i, '').strip
        permission = JSON.parse(Rxuniverse::API.cobol.run(:check_token, [token]))
        error! permission['error'], 403 unless permission['valid']
        @user = permission['user']
      end
    end

    route_definitions.each do |group|
      namespace group[:route], {desc: group[:desc]} do
        group[:methods].each do |mapping|
          desc mapping[:desc],
               entity: parse_response_entity([mapping[:method], group[:route], mapping[:route], 'Response'],
                                             mapping[:response]&.[](:members)),
               is_array: !!mapping[:response]&.[](:array)
          params do
            mapping[:params]&.reject { |parameter_definition| parameter_definition[:type].downcase == 'literal' }&.each do |parameter_definition|
              send(parameter_definition[:required] ? :requires : :optional,
                   parameter_definition[:name],
                   parse_parameter_definition(mapping, parameter_definition)
              )
            end
          end
          send(mapping[:method], mapping[:route]) do
            groups = Array(mapping[:params]).group_by { |param| !!param[:file] }
            file_params = Array(groups[true])
            normal_params = Array(groups[false])

            ordered_param_values = normal_params.map do |parameter_definition|
              Rxuniverse::API.extract_parameter_value(parameter_definition, params)
            end

            arg_file = nil
            begin
              if file_params.any?
                arg_file = Tempfile.new(%w(api .dat))
                file_params.each do |parameter_definition|
                  key = parameter_definition[:name]
                  value = Rxuniverse::API.extract_parameter_value(parameter_definition, params)
                  arg_file.write("#{key}=#{value}\n") if value
                end
                arg_file.flush
                ordered_param_values.push(arg_file.path)
              else
                ordered_param_values = [] if mapping[:params].length == 1 && ordered_param_values[0].nil? # if the route accepts an optional id and it has not been specified, pass null.  maybe clean this up?
              end
              data = Rxuniverse::API.cobol.run(mapping[:acu], ordered_param_values)
              result = JSON.parse(data)
            ensure
              arg_file&.close
              arg_file&.unlink
            end

            result
          end
        end
      end
    end

    desc "Redirect to swagger", hidden: true
    get :gui do
      redirect "/gui", permanent: true
    end

    desc 'Get info about the Rx-Universe™ API', hidden: true
    get :info do
      version = nil
      File.open(File.join(__dir__, '../package.json')) do |f|
        version = JSON.load(f, nil, create_additions: false, symbolize_names: true)&.[](:version)
      end
      {
          id: 'rx-universe-api',
          name: "Rx-Universe™ Lab Management Software API",
          description: "Application Programming Interface for the Rx-Universe Lab Management Software",
          useAuth: Rxuniverse::API.needs_authentication?,
          version: version,
      }
    end

    namespace :updates, desc: "Apply updates to Rx-Universe API" do
      desc "Show latest release version info", entity: parse_response_entity(['get', 'updates', 'status', 'Response'],
                                                                             [
                                                                                 {
                                                                                     "name": "current_version",
                                                                                     "desc": "The installed version of the API",
                                                                                     "type": "string"
                                                                                 },
                                                                                 {
                                                                                     "name": "latest_version",
                                                                                     "desc": "The installed version of the API",
                                                                                     "type": "string"
                                                                                 },
                                                                                 {
                                                                                     "name": "server",
                                                                                     "desc": "The configured update server - set this in the .env",
                                                                                     "type": "string"
                                                                                 },
                                                                                 {
                                                                                     "name": "contact_error",
                                                                                     "desc": "The error (if any) encountered trying to reach the update server for the latest update info",
                                                                                     "type": "string"
                                                                                 },
                                                                                 {
                                                                                     "name": "up_to_date",
                                                                                     "desc": "Denotes whether an update is ready to be applied",
                                                                                     "type": "boolean"
                                                                                 },
                                                                             ])
      get :check do
        contact_error = nil
        begin
          resp = Faraday.get("#{ENV['UPDATE_SERVER']}/packages")
          all_packages = JSON.parse(resp.body)
          applicable_packages = (all_packages['rx_universe_api'] || []).map do |v_str|
            begin
              Gem::Version.new(v_str)
            rescue StandardError => err
              nil
            end
          end.compact.sort
        rescue StandardError => err
          contact_error = err.message
          applicable_packages = []
        end

        latest_version = applicable_packages[-1]

        current_version = nil
        File.open(File.join(__dir__, '../package.json')) do |f|
          current_version = JSON.load(f, nil, create_additions: false, symbolize_names: true)[:version]
        end
        {
            current_version: current_version,
            latest_version: latest_version,
            server: ENV['UPDATE_SERVER'],
            contact_error: contact_error,
            up_to_date: !latest_version || (current_version == latest_version)
        }
      end

      desc "Upgrade to newest version - THIS WILL TAKE THE SYSTEM OFFLINE FOR SEVERAL MINUTES", entity: parse_response_entity(['post', 'updates', 'apply', 'Response'],
                                                                                                                              [
                                                                                                                                  {
                                                                                                                                      "name": "applying",
                                                                                                                                      "desc": "Whether or not a new update is currently being applied",
                                                                                                                                      "type": "boolean"
                                                                                                                                  },
                                                                                                                                  {
                                                                                                                                      "name": "error",
                                                                                                                                      "desc": "The error (if any) encountered trying to start the update process",
                                                                                                                                      "type": "string"
                                                                                                                                  },
                                                                                                                              ])
      post :apply do
        begin
          response = Faraday.get("#{ENV['UPDATE_SERVER']}/package/rx_universe_api")
          if response.status == 200
            update_folder = File.expand_path(File.join(__dir__, '..', 'storage', 'updates'))
            FileUtils.mkdir_p(update_folder) unless File.exists?(update_folder)
            update_file = File.create(File.join(update_folder, Time.now.strftime("%Y_%m_%d_%H_%M_%S")), 'w', binmode: true)
            update_file.write(response.body)
            update_file.flush
            update_file.close

            executable = File.join(RbConfig::CONFIG['bindir'], 'rake.cmd')
            arguments = [
                "apply_package",
                update_file.path
            ]
            Process.detach(spawn(%w(BUNDLER_ORIG_BUNDLER_VERSION BUNDLER_ORIG_BUNDLE_BIN_PATH BUNDLER_ORIG_BUNDLE_GEMFILE BUNDLER_ORIG_GEM_HOME BUNDLER_ORIG_GEM_PATH BUNDLER_ORIG_MANPATH BUNDLER_ORIG_PATH BUNDLER_ORIG_RB_USER_INSTALL BUNDLER_ORIG_RUBYLIB BUNDLER_ORIG_RUBYOPT BUNDLER_VERSION BUNDLE_BIN_PATH BUNDLE_GEMFILE GEM_HOME GEM_PATH RUBYLIB RUBYOPT).map do |k|
              [k, nil]
            end.to_h.merge({
                               'Path' => "#{RbConfig::CONFIG['bindir']};#{ENV['Path']}",
                               'TZ' => ENV['TZ']
                           }), executable, *arguments, chdir: Rails.root.to_s))
            {
                applying: true,
                error: nil
            }
          else
            {
                applying: false,
                error: 'Unreachable update server'
            }
          end
        end
      rescue StandardError => err
        {
            applying: false,
            error: err.message
        }
      end
    end

    add_swagger_documentation add_version: false,
                              mount_path: '/swagger',
                              doc_version: '1.0.0',
                              tags: [*Rxuniverse::API.route_definitions.map { |group| {name: group[:route], description: group[:desc]} }, {name: 'updates', description: "Apply updates to Rx-Universe API"}],
                              info: {
                                  title: "Rx-Universe™ LMS API",
                                  description: <<-DOCS
Interact with live Rx-Universe™ LMS Data by calling the endpoints below.
#{oauth2_instructions}
                                  DOCS
                              },
                              security_definitions: security_definitions,
                              security: security_object
  end
end
